"""
Admin site configurations for integrated channel's Content Metadata Transmission table.
"""

from django.contrib import admin

from integrated_channels.integrated_channel.models import ApiResponseRecord, ContentMetadataItemTransmission
from integrated_channels.utils import get_enterprise_customer_from_enterprise_enrollment


class BaseLearnerDataTransmissionAuditAdmin(admin.ModelAdmin):
    """
    Base admin class to hold commonly used methods across integrated channel admin views
    """
    def enterprise_customer_name(self, obj):
        """
        Returns: the name for the attached EnterpriseCustomer or None if customer object does not exist.
        Args:
            obj: The instance of Django model being rendered with this admin form.
        """
        ent_customer = get_enterprise_customer_from_enterprise_enrollment(obj.enterprise_course_enrollment_id)
        return ent_customer.name if ent_customer else None


@admin.action(description='Clear remote_deleted_at on ContentMetadataItemTransmission item(s)')
def clear_remote_deleted_at(modeladmin, request, queryset):  # pylint: disable=unused-argument
    queryset.update(remote_deleted_at=None)


@admin.register(ContentMetadataItemTransmission)
class ContentMetadataItemTransmissionAdmin(admin.ModelAdmin):
    """
    Admin for the ContentMetadataItemTransmission audit table
    """
    list_display = (
        'enterprise_customer',
        'integrated_channel_code',
        'content_id',
        'remote_deleted_at',
        'modified'
    )

    search_fields = (
        'enterprise_customer__name',
        'enterprise_customer__uuid',
        'integrated_channel_code',
        'content_id'
    )

    raw_id_fields = (
        'enterprise_customer',
    )

    readonly_fields = [
        'api_record',
        'api_response_status_code',
        'friendly_status_message',
    ]

    actions = [
        clear_remote_deleted_at,
    ]

    list_per_page = 1000


@admin.register(ApiResponseRecord)
class ApiResponseRecordAdmin(admin.ModelAdmin):
    """
    Admin for the ApiResponseRecord table
    """
    list_display = (
        'id',
        'status_code'
    )

    search_fields = (
        'id',
        'status_code'
    )

    readonly_fields = (
        'status_code',
        'body'
    )

    list_per_page = 1000



"""
Utilities common to different integrated channels.
"""

import base64
import itertools
import json
import math
import re
from datetime import datetime, timedelta
from itertools import islice
from logging import getLogger
from string import Formatter
from urllib.parse import urlparse

import pytz
import requests

from django.apps import apps
from django.core.exceptions import ObjectDoesNotExist
from django.db.models import Q
from django.utils import timezone
from django.utils.html import strip_tags

from enterprise.utils import parse_datetime_handle_invalid, parse_lms_api_datetime
from integrated_channels.catalog_service_utils import get_course_run_for_enrollment

UNIX_EPOCH = datetime(1970, 1, 1, tzinfo=timezone.utc)
UNIX_MIN_DATE_STRING = '1970-01-01T00:00:00Z'
UNIX_MAX_DATE_STRING = '2038-01-19T03:14:07Z'

LOGGER = getLogger(__name__)


def encode_data_for_logging(data):
    """
    Converts input into URL-safe, utf-8 encoded, base64 encoded output
    If the input is other than a string, it is dumped to json
    """
    if not isinstance(data, str):
        data = json.dumps(data)
    return base64.urlsafe_b64encode(data.encode("utf-8")).decode('utf-8')


def parse_datetime_to_epoch(datestamp, magnitude=1.0):
    """
    Convert an ISO-8601 datetime string to a Unix epoch timestamp in some magnitude.

    By default, returns seconds.
    """
    parsed_datetime = parse_lms_api_datetime(datestamp)
    time_since_epoch = parsed_datetime - UNIX_EPOCH
    return int(time_since_epoch.total_seconds() * magnitude)


def strip_html_tags(text, strip_entities=True):
    """
    Return (str): Text without any html tags and entities.

    Args:
        text (str): text having html tags
        strip_entities (bool): If set to True html entities are also stripped
    """
    text = strip_tags(text)
    if strip_entities:
        text = re.sub(r'&([a-zA-Z]{4,5}|#[0-9]{2,4});', '', text)
    return text


def parse_datetime_to_epoch_millis(datestamp):
    """
    Convert an ISO-8601 datetime string to a Unix epoch timestamp in milliseconds.
    """
    return parse_datetime_to_epoch(datestamp, magnitude=1000.0)


def current_time_is_in_interval(start, end):
    """
    Determine whether the current time is on the interval [start, end].
    """
    interval_start = parse_lms_api_datetime(start or UNIX_MIN_DATE_STRING)
    interval_end = parse_lms_api_datetime(end or UNIX_MAX_DATE_STRING)
    return interval_start <= timezone.now() <= interval_end


def chunks(dictionary, chunk_size):
    """
    Yield successive n-sized chunks from dictionary.
    """
    iterable = iter(dictionary)
    for __ in range(0, len(dictionary), chunk_size):
        yield {key: dictionary[key] for key in islice(iterable, chunk_size)}


def strfdelta(tdelta, fmt='{D:02}d {H:02}h {M:02}m {S:02}s', input_type='timedelta'):
    """
    Convert a datetime.timedelta object or a regular number to a custom-formatted string.

    This function works like the strftime() method works for datetime.datetime
    objects.

    The fmt argument allows custom formatting to be specified.  Fields can
    include seconds, minutes, hours, days, and weeks.  Each field is optional.

    Arguments:
        tdelta (datetime.timedelta, int): time delta object containing the duration or an integer
            to go with the input_type.
        fmt (str): Expected format of the time delta. place holders can only be one of the following.
            1. D to extract days from time delta
            2. H to extract hours from time delta
            3. M to extract months from time delta
            4. S to extract seconds from timedelta
        input_type (str):  The input_type argument allows tdelta to be a regular number instead of the
            default, which is a datetime.timedelta object.
            Valid input_type strings:
                1. 's', 'seconds',
                2. 'm', 'minutes',
                3. 'h', 'hours',
                4. 'd', 'days',
                5. 'w', 'weeks'
    Returns:
        (str): timedelta object interpolated into a string following the given format.

    Examples:
        '{D:02}d {H:02}h {M:02}m {S:02}s' --> '05d 08h 04m 02s' (default)
        '{W}w {D}d {H}:{M:02}:{S:02}'     --> '4w 5d 8:04:02'
        '{D:2}d {H:2}:{M:02}:{S:02}'      --> ' 5d  8:04:02'
        '{H}h {S}s'                       --> '72h 800s'
    """
    # Convert tdelta to integer seconds.
    if input_type == 'timedelta':
        remainder = int(tdelta.total_seconds())
    elif input_type in ['s', 'seconds']:
        remainder = int(tdelta)
    elif input_type in ['m', 'minutes']:
        remainder = int(tdelta) * 60
    elif input_type in ['h', 'hours']:
        remainder = int(tdelta) * 3600
    elif input_type in ['d', 'days']:
        remainder = int(tdelta) * 86400
    elif input_type in ['w', 'weeks']:
        remainder = int(tdelta) * 604800
    else:
        raise ValueError(
            'input_type is not valid. Valid input_type strings are: "timedelta", "s", "m", "h", "d", "w"'
        )

    f = Formatter()
    desired_fields = [field_tuple[1] for field_tuple in f.parse(fmt)]
    possible_fields = ('W', 'D', 'H', 'M', 'S')
    constants = {'W': 604800, 'D': 86400, 'H': 3600, 'M': 60, 'S': 1}
    values = {}

    for field in possible_fields:
        if field in desired_fields and field in constants:
            values[field], remainder = divmod(remainder, constants[field])

    return f.format(fmt, **values)


def convert_comma_separated_string_to_list(comma_separated_string):
    """
    Convert the comma separated string to a valid list.
    """
    return list({item.strip() for item in comma_separated_string.split(",") if item.strip()})


def get_image_url(content_metadata_item):
    """
    Return the image URI of the content item.
    """
    image_url = ''
    if content_metadata_item['content_type'] == 'program':
        image_url = content_metadata_item.get('card_image_url')
    elif content_metadata_item['content_type'] in ['course', 'courserun']:
        image_url = content_metadata_item.get('image_url')

    return image_url


def is_already_transmitted(
    transmission,
    enterprise_enrollment_id,
    enterprise_configuration_id,
    grade,
    subsection_id=None,
    detect_grade_updated=True,
):
    """
    Returns: Boolean indicating if completion date for given enrollment is already sent of not.

    Args:
        transmission: TransmissionAudit model to search enrollment in
        enterprise_enrollment_id: enrollment id
        grade: 'Pass' or 'Fail' status
        subsection_id (Optional): The id of the subsection, needed if transmitting assessment level grades as there can
        be multiple per course.
        detect_grade_updated: default True. if this is False, method does not take into account grade changes
    """
    try:
        already_transmitted = transmission.objects.filter(
            enterprise_course_enrollment_id=enterprise_enrollment_id,
            plugin_configuration_id=enterprise_configuration_id,
            error_message='',
            status__lt=400
        )
        if subsection_id:
            already_transmitted = already_transmitted.filter(subsection_id=subsection_id)

        latest_transmitted_tx = already_transmitted.latest('id')
        if detect_grade_updated:
            return latest_transmitted_tx and getattr(latest_transmitted_tx, 'grade', None) == grade
        return latest_transmitted_tx is not None
    except transmission.DoesNotExist:
        return False


def get_duration_from_estimated_hours(estimated_hours):
    """
    Return the duration in {hours}:{minutes}:00 corresponding to estimated hours as int or float.
    """
    if estimated_hours and isinstance(estimated_hours, (int, float)):
        fraction, whole_number = math.modf(estimated_hours)
        hours = "{:02d}".format(int(whole_number))
        minutes = "{:02d}".format(int(60 * fraction))
        duration = "{hours}:{minutes}:00".format(hours=hours, minutes=minutes)
        return duration

    return None


def get_courserun_duration_in_hours(course_run):
    """
    Return the approximate number of hours required to complete this course run.
    """
    min_effort = course_run.get('min_effort')
    max_effort = course_run.get('max_effort')
    weeks_to_complete = course_run.get('weeks_to_complete')
    if min_effort and max_effort and weeks_to_complete:
        average_hours_per_week = (min_effort + max_effort) / 2.0
        total_hours = weeks_to_complete * average_hours_per_week
        return math.ceil(total_hours)
    else:
        return 0


def get_subjects_from_content_metadata(content_metadata_item):
    """
    Returns a list of subject names for the content metadata item.

    Subjects in the content metadata item are represented by either:
      - a list of strings, e.g. ['Communication']
      - a list of objects, e.g. [{'name': 'Communication'}]

    Arguments:
        - content_metadata_item (dict): a dictionary for the content metadata item

    Returns:
        - list: a list of subject names as strings
    """
    metadata_subjects = content_metadata_item.get('subjects') or []
    subjects = set()

    for subject in metadata_subjects:
        if isinstance(subject, str):
            subjects.add(subject)
            continue

        subject_name = subject.get('name')
        if subject_name:
            subjects.add(subject_name)

    return list(subjects)


def generate_formatted_log(
    channel_name=None,
    enterprise_customer_uuid=None,
    lms_user_id=None,
    course_or_course_run_key=None,
    message=None,
    plugin_configuration_id=None,
):
    """
    Formats and returns a standardized message for the integrated channels.
    All fields are optional.

    Arguments:
    - channel_name (str): The name of the integrated channel
    - enterprise_customer_uuid (str): UUID of the relevant EnterpriseCustomer
    - lms_user_id (str): The LMS User id (if applicable) related to the message
    - course_or_course_run_key (str): The course key (if applicable) for the message
    - message (str): The string to be formatted and logged
    - plugin_configuration_id (str): The configuration id related to the message

    """
    return f'integrated_channel={channel_name}, '\
        f'integrated_channel_enterprise_customer_uuid={enterprise_customer_uuid}, '\
        f'integrated_channel_lms_user={lms_user_id}, '\
        f'integrated_channel_course_key={course_or_course_run_key}, '\
        f'integrated_channel_plugin_configuration_id={plugin_configuration_id}, {message}'


def refresh_session_if_expired(
    oauth_access_token_function,
    session=None,
    expires_at=None,
):
    """
    Instantiate a new session object for use in connecting with integrated channel.
    Or, return an updated session if provided session has expired.
    Suitable for use with oauth supporting servers that use bearer token: Canvas, Blackboard etc.

    Arguments:
        - oauth_access_token_function (function): access token fetch function
        - session (requests.Session): a session object. Pass None if creating new session
        - expires_at: the expiry date of the session if known. None is interpreted as expired.

    Each enterprise customer connecting to channel should have a single client session.
    Will only create a new session if token expiry has been reached
    If a new session is being created, closes the session first

    Returns (tuple) with values:
     - session: newly created session or an updated session (can be stored for later use)
     - expires_at: new expiry date to be stored for later use
    If session has not expired, or not updated for any reason, just returns the input values of
    session and expires_at
    """
    now = datetime.utcnow()
    if session is None or expires_at is None or now >= expires_at:
        # need new session if session expired, or not initialized
        if session:
            session.close()
        # Create a new session with a valid token
        oauth_access_token, expires_in = oauth_access_token_function()
        new_session = requests.Session()
        new_session.headers['Authorization'] = 'Bearer {}'.format(oauth_access_token)
        new_session.headers['content-type'] = 'application/json'
        # expiry expected after `expires_in` seconds
        if expires_in is not None:
            new_expires_at = datetime.utcnow() + timedelta(seconds=expires_in)
        return new_session, new_expires_at
    return session, expires_at


def get_upgrade_deadline(course_run):
    """
    Returns upgrade_deadline of a verified seat if found. Otherwise returns None.
    """
    for seat in course_run.get('seats', []):
        if seat.get('type') == 'verified':
            return parse_datetime_handle_invalid(seat.get('upgrade_deadline'))
    return None


def is_course_completed(enterprise_enrollment, is_passing, incomplete_count, passed_timestamp=None):
    '''
    For non audit, this requires passing and passed_timestamp
    For audit enrollment, returns True if:
     - for non upgradable course:
        - no more non-gated content is left
     - for upgradable course:
        - the verified upgrade deadline has passed AND no more non-gated content is left
    '''
    if enterprise_enrollment.is_audit_enrollment:
        if incomplete_count is None:
            raise ValueError('Incomplete count is required if using audit enrollment')
        course_run = get_course_run_for_enrollment(enterprise_enrollment)
        upgrade_deadline = get_upgrade_deadline(course_run)
        if upgrade_deadline is None:
            return incomplete_count == 0
        else:
            # for upgradable course check deadline passed as well
            now = datetime.now(pytz.UTC)
            return incomplete_count == 0 and upgrade_deadline < now
    return passed_timestamp is not None and is_passing


def is_valid_url(url):
    """
    Return where the specified URL is a valid absolute url.
    """
    if len(url) == 0:
        return True
    try:
        result = urlparse(url)
        return all([result.scheme, result.netloc])
    except ValueError:
        return False


def batch_by_pk(ModelClass, extra_filter=Q(), batch_size=10000):
    """
    yield per batch efficiently
    using limit/offset does a lot of table scanning to reach higher offsets
    this scanning can be slow on very large tables
    if you order by pk, you can use the pk as a pivot rather than offset
    this utilizes the index, which is faster than scanning to reach offset
    Example usage:
    csod_only_filter = Q(integrated_channel_code='CSOD')
    for items_batch in batch_by_pk(ContentMetadataItemTransmission, extra_filter=csod_only_filter):
        for item in items_batch:
            ...
    """
    qs = ModelClass.objects.filter(extra_filter).order_by('pk')[:batch_size]
    while qs.exists():
        yield qs
        # qs.last() doesn't work here because we've already sliced
        # loop through so we eventually grab the last one
        for item in qs:
            start_pk = item.pk
        qs = ModelClass.objects.filter(pk__gt=start_pk).filter(extra_filter).order_by('pk')[:batch_size]


def truncate_item_dicts(items_to_create, items_to_update, items_to_delete, combined_maximum_size):
    """
    given the item collections to create, update, and delete on the remote LMS side, truncate the
    collections to keep under a maximum batch size, prioritizing creates, updates, then deletes
    """
    # if we have more to work with than the allowed space, slice it up
    if len(items_to_create) + len(items_to_delete) + len(items_to_update) > combined_maximum_size:
        # prioritize creates, then updates, then deletes
        items_to_create = dict(itertools.islice(items_to_create.items(), combined_maximum_size))
        count_left = combined_maximum_size - len(items_to_create)
        items_to_update = dict(itertools.islice(items_to_update.items(), count_left))
        count_left = count_left - len(items_to_update)
        items_to_delete = dict(itertools.islice(items_to_delete.items(), count_left))
    return items_to_create, items_to_update, items_to_delete


def channel_code_to_app_label(channel_code):
    """
    Convert an integrated_channel channel_code to app_label.
    They can be different, such as in the case of SAP -> sap_success_factors
    """
    app_label = channel_code.lower()
    if app_label == 'generic':
        app_label = 'integrated_channel'
    elif app_label == 'sap':
        app_label = 'sap_success_factors'
    elif app_label == 'csod':
        app_label = 'cornerstone'
    return app_label


def get_enterprise_customer_model():
    """
    Returns the ``EnterpriseCustomer`` class.
    """
    return apps.get_model('enterprise', 'EnterpriseCustomer')


def get_enterprise_customer_from_enterprise_enrollment(enrollment_id):
    """
    Returns the Django ORM enterprise customer object that is associated with an enterprise enrollment ID
    """
    EnterpriseCustomer = get_enterprise_customer_model()
    try:
        ec = EnterpriseCustomer.objects.filter(
            enterprise_customer_users__enterprise_enrollments=enrollment_id
        ).first()
        return ec
    except ObjectDoesNotExist:
        return None


def get_enterprise_client_by_channel_code(channel_code):
    """
    Get the appropriate enterprise client based on channel code
    """
    # TODO: Other configs
    from integrated_channels.canvas.client import CanvasAPIClient  # pylint: disable=C0415
    _enterprise_client_model_by_channel_code = {
        'canvas': CanvasAPIClient,
    }
    return _enterprise_client_model_by_channel_code[channel_code]



# Generated by Django 3.2.12 on 2022-03-02 22:31

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('moodle', '0010_auto_20220131_1539'),
    ]

    operations = [
        migrations.AlterField(
            model_name='historicalmoodleenterprisecustomerconfiguration',
            name='display_name',
            field=models.CharField(blank=True, default='', help_text='A configuration nickname.', max_length=255),
        ),
        migrations.AlterField(
            model_name='moodleenterprisecustomerconfiguration',
            name='display_name',
            field=models.CharField(blank=True, default='', help_text='A configuration nickname.', max_length=255),
        ),
    ]



# Generated by Django 3.2.15 on 2022-09-29 19:01

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0163_auto_20220928_1611'),
        ('integrated_channel', '0021_remove_contentmetadataitemtransmission_api_response_body'),
    ]

    operations = [
        migrations.AlterIndexTogether(
            name='contentmetadataitemtransmission',
            index_together={('enterprise_customer', 'integrated_channel_code', 'plugin_configuration_id', 'content_id')},
        ),
    ]



# Generated by Django 3.2.12 on 2022-03-15 12:15

from django.conf import settings
from django.db import migrations
from django.db.models import Count, Min


def remove_duplicate_entries(apps, schema_editor):
    SystemWideEnterpriseUserRoleAssignment = apps.get_model('enterprise', 'SystemWideEnterpriseUserRoleAssignment')

    duplicates = (
        SystemWideEnterpriseUserRoleAssignment.objects.values('user_id', 'role_id', 'enterprise_customer_id')
            .annotate(first_entry_id=Min('id'), num_entries=Count('id'))
            .filter(num_entries__gt=1)
    )

    # delete the duplicates, excluding the first role assignment for the user/role/enterprise combination
    for duplicate in duplicates:
        (
            SystemWideEnterpriseUserRoleAssignment.objects
                .filter(
                    user_id=duplicate['user_id'],
                    role_id=duplicate['role_id'],
                    enterprise_customer_id=duplicate['enterprise_customer_id']
                )
                .exclude(id=duplicate['first_entry_id'])
                .delete()
        )


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('enterprise', '0153_add_enable_browse_and_request'),
    ]

    operations = [
        migrations.RunPython(remove_duplicate_entries, reverse_code=migrations.RunPython.noop),
        migrations.AlterUniqueTogether(
            name='systemwideenterpriseuserroleassignment',
            unique_together={('enterprise_customer', 'role', 'user')},
        ),
    ]



"""
Management command for making things with test factories
"""

import logging
import re
import sys

import factory
from factory.declarations import SubFactory

from django.core.exceptions import ImproperlyConfigured
from django.core.management.base import BaseCommand, CommandError, SystemCheckError, handle_default_options
from django.db import connections

# We have to import the enterprise test factories to ensure it's loaded and found by __subclasses__
# To ensure factories outside of the enterprise package are loaded and found by the script,
# add any additionally desired factories as an import to this file. Make sure to catch the ImportError
# incase other consumers of the command do not have the same factories installed.
# For example:
try:
    import common.djangoapps.student.tests.factories  # pylint: disable=unused-import

    from test_utils import factories  # pylint: disable=unused-import
except ImportError:
    pass

log = logging.getLogger(__name__)


def convert_to_pascal(string):
    """
    helper method to convert strings to Pascal case.
    """
    return string.replace("_", " ").title().replace(" ", "")


def pairwise(iterable):
    """
    Convert a list into a list of tuples of adjacent elements.
    s -> [ (s0, s1), (s2, s3), (s4, s5), ... ]
    """
    a = iter(iterable)
    return zip(a, a)


def all_subclasses(cls):
    """
    Recursively get all subclasses of a class
    https://stackoverflow.com/a/3862957
    """
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)])


def convert_to_snake(string):
    """
    Helper method to convert strings to snake case.
    """
    return re.sub(r'(?<!^)(?=[A-Z])', '_', string).lower()


class Node():
    """
    Non-binary tree node class for building out a dependency tree of objects to create with customizations.
    """
    def __init__(self, data):
        self.data = data
        self.children = []
        self.customizations = {}
        self.factory = None
        self.instance = None

    def set_single_customization(self, field, value):
        """
        Set a single customization value to the current node, overrides existing values under the same key.
        """
        self.customizations[field] = value

    def add_child(self, obj):
        """
        Add a child to the current node
        """
        self.children.append(obj)

    def find_value(self, value):
        """
        Find a value in the tree
        """
        if self.data == value:
            return self
        else:
            for child in self.children:
                found = child.find_value(value)
                if found:
                    return found
            return None

    def build_records(self):
        """
        Recursively build out the tree of objects by first dealing with children nodes before getting to the parent.
        """
        built_children = {}
        for child in self.children:
            # if we have an instance, use it instead of creating more objects
            if child.instance:
                built_children.update({convert_to_snake(child.data): child.instance})
            else:
                # Use the output of child ``build_records`` to create the current level.
                built_child = child.build_records()
                built_children.update(built_child)

        # The data factory kwargs are specified custom fields + the PK's of generated child objects
        object_fields = self.customizations.copy()
        object_fields.update(built_children)

        # Some edge case sanity checking
        if not self.factory:
            raise CommandError(f"Cannot build objects as {self} does not have a factory")

        built_object = self.factory(**object_fields)
        object_data = {convert_to_snake(self.data): built_object}
        return object_data

    def __str__(self, level=0):
        """
        Overridden str method to allow for proper tree printing
        """
        if self.instance:
            body = f"PK: {self.instance.pk}"
        else:
            body = f"fields: {self.customizations}"
        ret = ("\t" * level) + f"{repr(self.data)} {body}" + "\n"
        for child in self.children:
            ret += child.__str__(level + 1)
        return ret

    def __repr__(self):
        """
        Overridden repr
        """
        return f'<Tree Node {self.data}>'


def build_tree_from_field_list(list_of_fields, provided_factory, base_node, customization_value):
    """
    Builds a non-binary tree of nodes based on a list of children nodes, using a base node and it's associated data
    factory as the parent node the user provided value as a reference to a potential, existing record.

    - list_of_fields (list of strings): the linked list of associated objects to create. Example-
        ['enterprise_customer_user', 'enterprise_customer', 'site']
    - provided_factory (factory.django.DjangoModelFactory): The data factory of the base_node.
    - base_node (Node): The parent node of the desired tree to build.
    - customization_value (string): The value to be assigned to the object associated with the last value in the
        ``list_of_fields`` param. Can either be a FK if the last value is a subfactory, or alternatively
        a custom value to be assigned to the field. Example-
        list_of_fields = ['enterprise_customer_user', 'enterprise_customer', 'site'],
        customization_value = 9
        or
        list_of_fields = ['enterprise_customer_user', 'enterprise_customer', 'name'],
        customization_value = "FRED"
    """
    current_factory = provided_factory
    current_node = base_node
    for index, value in enumerate(list_of_fields):
        try:
            # First we need to figure out if the current field is a sub factory or not
            f = getattr(current_factory, value)
            if isinstance(f, SubFactory):
                fk_object = None
                f_model = f.get_factory()._meta.get_model_class()

                # if we're at the end of the list
                if index == len(list_of_fields) - 1:
                    # verify that the provided customization value is a valid pk for the model
                    try:
                        fk_object = f_model.objects.get(pk=customization_value)
                    except f_model.DoesNotExist as exc:
                        raise CommandError(
                            f"Provided FK value: {customization_value} does not exist on {f_model.__name__}"
                        ) from exc

                # Look for the node in the tree
                if node := current_node.find_value(f_model.__name__):
                    # Not supporting customizations and FK's
                    if (bool(node.customizations) or bool(node.children)) and bool(fk_object):
                        raise CommandError("This script does not support customizing provided existing objects")
                    # If we found the valid FK earlier, assign it to the node
                    if fk_object:
                        node.instance = fk_object
                    # Add the field to the children of the current node
                    if node not in current_node.children:
                        current_node.add_child(node)
                    # Set current node and move on
                    current_node = node
                else:
                    # Create a new node
                    node = Node(
                        f_model.__name__,
                    )
                    node.factory = f.get_factory()
                    # If we found the valid FK earlier, assign it to the node
                    if fk_object:
                        node.instance = fk_object
                    # Add the field to the children of the current node
                    current_node.add_child(node)

                current_node = node
                current_factory = f.get_factory()
            else:
                if current_node.instance:
                    raise CommandError("This script cannot modify existing objects")
                current_node.set_single_customization(value, customization_value)
        except AttributeError as exc:
            log.error(f'Could not find value: {value} in factory: {current_factory}')
            raise CommandError(f'Could not find value: {value} in factory: {current_factory}') from exc
    return base_node


class Command(BaseCommand):
    """
    Management command for generating Django records from factories with custom attributes

    Example usage:
        $ ./manage.py manufacture_data --model enterprise.models.enterprise_customer \
            --name "Test Enterprise" --slug "test-enterprise"
    """

    def add_arguments(self, parser):
        parser.add_argument(
            '--model',
            dest='model',
            help='The model for which the record will be written',
        )

    def run_from_argv(self, argv):
        """
        Re-implemented from https://github.com/django/django/blob/main/django/core/management/base.py#L395 in order to
        support individual field customization. We will need to keep this method up to date with our current version of
        Django BaseCommand.

        Uses ``parse_known_args`` instead of ``parse_args`` to not throw an error when encountering unknown arguments

        https://docs.python.org/3.8/library/argparse.html#argparse.ArgumentParser.parse_known_args
        """
        self._called_from_command_line = True
        parser = self.create_parser(argv[0], argv[1])
        options, unknown = parser.parse_known_args(argv[2:])

        # Add the unknowns into the options for use of the handle method
        paired_unknowns = pairwise(unknown)
        field_customizations = {}
        for field, value in paired_unknowns:
            field_customizations[field.strip("--")] = value
        options.field_customizations = field_customizations

        cmd_options = vars(options)
        # Move positional args out of options to mimic legacy optparse
        args = cmd_options.pop("args", ())
        handle_default_options(options)
        try:
            self.execute(*args, **cmd_options)
        except CommandError as e:
            if options.traceback:
                raise

            # SystemCheckError takes care of its own formatting.
            if isinstance(e, SystemCheckError):
                self.stderr.write(str(e), lambda x: x)
            else:
                self.stderr.write("%s: %s" % (e.__class__.__name__, e))
            sys.exit(e.returncode)
        finally:
            try:
                connections.close_all()
            except ImproperlyConfigured:
                # Ignore if connections aren't setup at this point (e.g. no
                # configured settings).
                pass

    def handle(self, *args, **options):
        """
        Entry point for management command execution.
        """
        if not options.get('model'):
            log.error("Did not receive a model")
            raise CommandError("Did not receive a model")

        # Convert to Pascal case if the provided name is snake case/is all lowercase
        path_of_model = options.get('model').split(".")
        if '_' in path_of_model[-1] or path_of_model[-1].islower():
            last_path = convert_to_pascal(path_of_model[-1])
        else:
            last_path = path_of_model[-1]

        provided_model = '.'.join(path_of_model[:-1]) + '.' + last_path
        # Get all installed/imported factories
        factories_list = all_subclasses(factory.django.DjangoModelFactory)
        # Find the factory that matches the provided model
        for potential_factory in factories_list:
            # Fetch the model for the factory
            factory_model = potential_factory._meta.model
            # Check if the factories model matches the provided model
            if f"{factory_model.__module__}.{factory_model.__name__}" == provided_model:
                # Now that we have the right factory, we can build according to the provided custom attributes
                field_customizations = options.get('field_customizations', {})
                base_node = Node(factory_model.__name__)
                base_node.factory = potential_factory
                # For each provided custom attribute...
                for field, value in field_customizations.items():

                    # We need to build a tree of objects to be created and may be customized by other custom attributes
                    stripped_field = field.strip("--")
                    fk_field_customization_split = stripped_field.split("__")
                    base_node = build_tree_from_field_list(
                        fk_field_customization_split,
                        potential_factory,
                        base_node,
                        value,
                    )

                built_node = base_node.build_records()
                log.info(f"\nGenerated factory data: \n{base_node}")
                return str(list(built_node.values())[0].pk)

        log.error(f"Provided model: {provided_model} does not exist or does not have an associated factory")
        raise CommandError(f"Provided model: {provided_model}'s factory is not imported or does not exist")



# Generated by Django 2.2.24 on 2021-12-03 18:00

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0150_invite_key_expiry_required'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomerinvitekey',
            name='is_active',
            field=models.BooleanField(default=True, help_text='Specifies if the key is active. Once deactivated, the key is no longer valid and cannot be reactivated.'),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomerinvitekey',
            name='is_active',
            field=models.BooleanField(default=True, help_text='Specifies if the key is active. Once deactivated, the key is no longer valid and cannot be reactivated.'),
        ),
    ]



"""
Backfill missing audit record foreign keys.
"""
import logging
from datetime import datetime, timedelta

from django.core.management.base import BaseCommand
from django.utils.translation import gettext as _

from integrated_channels.blackboard.models import (
    BlackboardEnterpriseCustomerConfiguration,
    BlackboardLearnerAssessmentDataTransmissionAudit,
    BlackboardLearnerDataTransmissionAudit,
)
from integrated_channels.canvas.models import (
    CanvasEnterpriseCustomerConfiguration,
    CanvasLearnerAssessmentDataTransmissionAudit,
    CanvasLearnerDataTransmissionAudit,
)
from integrated_channels.cornerstone.models import (
    CornerstoneEnterpriseCustomerConfiguration,
    CornerstoneLearnerDataTransmissionAudit,
)
from integrated_channels.degreed2.models import (
    Degreed2EnterpriseCustomerConfiguration,
    Degreed2LearnerDataTransmissionAudit,
)
from integrated_channels.degreed.models import (
    DegreedEnterpriseCustomerConfiguration,
    DegreedLearnerDataTransmissionAudit,
)
from integrated_channels.integrated_channel.management.commands import IntegratedChannelCommandMixin
from integrated_channels.integrated_channel.models import (
    ContentMetadataItemTransmission,
    GenericEnterpriseCustomerPluginConfiguration,
    GenericLearnerDataTransmissionAudit,
)
from integrated_channels.moodle.models import MoodleEnterpriseCustomerConfiguration, MoodleLearnerDataTransmissionAudit
from integrated_channels.sap_success_factors.models import (
    SAPSuccessFactorsEnterpriseCustomerConfiguration,
    SapSuccessFactorsLearnerDataTransmissionAudit,
)

MODELS = {
    'MOODLE': [MoodleEnterpriseCustomerConfiguration, MoodleLearnerDataTransmissionAudit],
    'CSOD': [CornerstoneEnterpriseCustomerConfiguration, CornerstoneLearnerDataTransmissionAudit],
    'BLACKBOARD': [BlackboardEnterpriseCustomerConfiguration, BlackboardLearnerDataTransmissionAudit],
    'BLACKBOARD_ASMT': [BlackboardEnterpriseCustomerConfiguration, BlackboardLearnerAssessmentDataTransmissionAudit],
    'CANVAS': [CanvasEnterpriseCustomerConfiguration, CanvasLearnerDataTransmissionAudit],
    'CANVAS_ASMT': [CanvasEnterpriseCustomerConfiguration, CanvasLearnerAssessmentDataTransmissionAudit],
    'DEGREED': [DegreedEnterpriseCustomerConfiguration, DegreedLearnerDataTransmissionAudit],
    'DEGREED2': [Degreed2EnterpriseCustomerConfiguration, Degreed2LearnerDataTransmissionAudit],
    'GENERIC': [GenericEnterpriseCustomerPluginConfiguration, GenericLearnerDataTransmissionAudit],
    'SAP': [SAPSuccessFactorsEnterpriseCustomerConfiguration, SapSuccessFactorsLearnerDataTransmissionAudit],
}

LOGGER = logging.getLogger(__name__)


class Command(IntegratedChannelCommandMixin, BaseCommand):
    """
    Management command which backfills missing audit record foreign keys.
    """
    help = _('''
    Set error state for configurations.
    ''')

    def update_config_last_errored_at(self):
        """
        For each audit record kind (learner and content), find all the records in batch, and lookup
        if they've had recent sync errors in the last day. If not, clear out the last_content_sync_errored_at
        value associated with the configuration.
        """
        try:
            has_learner_errors, has_content_errors = True, True
            yesterday = datetime.utcnow() - timedelta(days=1)
            for channel_code, (ConfigModel, LearnerAuditModel) in MODELS.items():
                configs = ConfigModel.objects.all()
                for config in configs:
                    if config.last_sync_errored_at is None:
                        continue
                    customer = config.enterprise_customer
                    plugin_id = config.id
                    # learner audits
                    errored_learner_audits = LearnerAuditModel.objects.filter(
                        created__date__gt=yesterday,
                        status__gt=299,
                        enterprise_customer_uuid=customer.uuid,
                        plugin_configuration_id=plugin_id,
                    )
                    if not errored_learner_audits:
                        config.last_learner_sync_errored_at = None
                        has_learner_errors = False
                    # content metadata audits
                    errored_content_audits = ContentMetadataItemTransmission.objects.filter(
                        remote_created_at__gt=yesterday,
                        enterprise_customer=customer,
                        integrated_channel_code=channel_code,
                        plugin_configuration_id=plugin_id,
                        api_response_status_code__gt=299
                    )
                    if not errored_content_audits:
                        config.last_content_sync_errored_at = None
                        has_content_errors = False
                    if not has_learner_errors and not has_content_errors:
                        config.last_sync_errored_at = None
                    if not has_learner_errors or not has_content_errors:
                        LOGGER.info(
                            'Config with id {}, channel code {}, enterprise customer {}'
                            ' error information has been updated'.format(
                                config.id, channel_code, config.enterprise_customer.uuid
                            )
                        )
                    config.save()
        except Exception as exc:
            LOGGER.exception('update_config_last_errored_at', exc_info=exc)
            raise exc

    def handle(self, *args, **options):
        """
        Set error state for configurations.
        """
        LOGGER.info('Begin nulling out outdated last_sync_errored_at in configs')
        self.update_config_last_errored_at()
        LOGGER.info('Finished nulling out outdated last_sync_errored_at in configs')



# Generated by Django 3.2.16 on 2022-12-20 15:27

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('moodle', '0021_move_and_recrete_completed_timestamp'),
    ]

    operations = [
        migrations.AddField(
            model_name='historicalmoodleenterprisecustomerconfiguration',
            name='grade_assignment_name',
            field=models.CharField(default='(edX integration) Final Grade', help_text='The name for the grade assigment created for the grade integration.', max_length=255, verbose_name='Grade Assignment Name'),
        ),
        migrations.AddField(
            model_name='historicalmoodleenterprisecustomerconfiguration',
            name='grade_scale',
            field=models.IntegerField(default=100, help_text='The maximum grade points for the courses. Default: 100', verbose_name='Grade Scale'),
        ),
        migrations.AddField(
            model_name='moodleenterprisecustomerconfiguration',
            name='grade_assignment_name',
            field=models.CharField(default='(edX integration) Final Grade', help_text='The name for the grade assigment created for the grade integration.', max_length=255, verbose_name='Grade Assignment Name'),
        ),
        migrations.AddField(
            model_name='moodleenterprisecustomerconfiguration',
            name='grade_scale',
            field=models.IntegerField(default=100, help_text='The maximum grade points for the courses. Default: 100', verbose_name='Grade Scale'),
        ),
    ]



"""
Module for defining models needed for configuration of internal
concerns like management command options and parameters.
"""
from config_models.models import ConfigurationModel

from django.db import models
from django.utils.translation import gettext_lazy as _

from enterprise.constants import (
    ENTERPRISE_ADMIN_ROLE,
    ENTERPRISE_LEARNER_ROLE,
    ENTERPRISE_OPERATOR_ROLE,
    SYSTEM_ENTERPRISE_CATALOG_ADMIN_ROLE,
)


# This is not a feature toggle.
class UpdateRoleAssignmentsWithCustomersConfig(ConfigurationModel):
    """
    Model that specifies parameters for the
    ``update_role_assignments_with_customers`` management command.

    .. no_pii:
    """
    class Meta:
        app_label = 'enterprise'

    ROLE_CHOICES = [
        (ENTERPRISE_ADMIN_ROLE, ENTERPRISE_ADMIN_ROLE),
        (ENTERPRISE_LEARNER_ROLE, ENTERPRISE_LEARNER_ROLE),
        (SYSTEM_ENTERPRISE_CATALOG_ADMIN_ROLE, SYSTEM_ENTERPRISE_CATALOG_ADMIN_ROLE),
        (ENTERPRISE_OPERATOR_ROLE, ENTERPRISE_OPERATOR_ROLE),
    ]
    role = models.CharField(
        blank=True,
        choices=ROLE_CHOICES,
        max_length=100,
        help_text=_('Specifies which user role assignments to update.  If unspecified, will update for all roles.'),
    )
    batch_size = models.IntegerField(
        default=500,
        help_text=_('Number of user role asssignments to update in each batch of updates.'),
    )
    enterprise_customer_uuid = models.CharField(
        blank=True,
        max_length=36,
        help_text=_('The enterprise customer to limit role assignments to.'),
    )
    dry_run = models.BooleanField(
        default=True,
        help_text=_(
            'If set, no updates or creates will occur; will instead iterate over '
            'the assignments that would be modified or created'
        ),
    )
